'use strict';

const pkg = require( '../package' );
const Config = require( '@simple-emotion/config' )( pkg.name );
const Log = require( '@simple-emotion/log' );
const Validate = require( '@simple-emotion/validate-struct' );
const CloudStorage = require( '@simple-emotion/cloud-storage' );
const Utils = require( './utils' );
const Schemas = require( './schemas' );
const fs = require( 'fs' );
const path = require( 'path' );
const uuid = require( 'uuid' );
const amqp = require( 'amqplib/callback_api' );

const default_config = require( `../configs/app.json` );

if ( !( 'toJSON' in Error.prototype ) ) {
  Object.defineProperty(
    Error.prototype,
    'toJSON',
    {
      value: function () {

        const alt = {};

        Object.getOwnPropertyNames( this ).forEach( key => {
          alt[ key ] = this[ key ];
        }, this );

        return alt;

      },
      configurable: true,
      writable: true
    }
  );
}

class TranscodeWorker {

  constructor( config ) {
    this.config = Config.load( default_config, config, pkg.name );
    this.worker = this.config.workers[ pkg.name ];
    this.log = Log( pkg.name, pkg.version, Config.current );
    this.cloudStorage = new CloudStorage( this.config.services );
    this.running = false;
  }

  start( cb ) {

    if ( this.running ) {
      return;
    }

    this.running = true;

    const storage_dir = path.resolve( this.worker.paths.storage );

    // Ensure storage dir exists
    fs.access( storage_dir, fs.constants.R_OK, err => {

      if ( err ) {

        this.running = false;

        const error = {
          reason: err.message,
          message: 'Storage directory not found.'
        };

        this.log.error( error );

        return cb( error );

      }

      // Clear the storage dir
      Utils.rmdir( `${storage_dir}/*`, err => {

        if ( err ) {
          this.running = false;
          this.log.error( err );
          return cb( err );
        }

        amqp.connect( this.config.services.rabbitmq.url, ( err, conn ) => {

          if ( err ) {
            this.running = false;
            this.log.error( err );
            return cb( err );
          }

          conn.on( 'error', err => {
            this.log.error( '[AMQP] Connection error.' );
            this.log.error( err );
            process.exit( 1 );
          } );

          conn.on( 'close', () => {
            this.log.error( '[AMQP] Connection closed.' );
            process.exit( 1 );
          } );

          conn.createChannel( ( err, ch ) => {

            if ( err ) {
              this.running = false;
              this.log.error( err );
              return cb( err );
            }

            ch.on( 'error', err => {
              this.log.error( '[AMQP] Channel error.' );
              this.log.error( err );
              process.exit( 1 );
            } );

            ch.on( 'close', () => {
              this.log.error( '[AMQP] Channel closed.' );
              process.exit( 1 );
            } );

            ch.prefetch( this.worker.prefetch );

            ch.consume(
              this.worker.queues.request,
              this.__handleConsume( ch ).bind( this ),
              {
                noAck: false
              }
            );

          } );

        } );

      } );

    } );

  }

  __handleConsume( ch ) {
    return msg => {

      // TODO: DO NOT RETRY INDEFINITELY BY CHECKING DEATH COUNT

      let payload;

      try {
        payload = JSON.parse( msg.content.toString() );
      }
      catch ( err ) {
        return this.__respond(
          ch,
          msg,
          {
            error: {
              code: 400,
              message: 'Received malformed message.',
              reason: err.message
            }
          }
        );
      }

      Validate( payload, { type: Schemas.request }, 'request', ( err, data ) => {

        if ( err ) {
          return this.__respond(
            ch,
            msg,
            {
              error: {
                code: 400,
                message: 'Received invalid message.',
                reason: err.message
              },
              metadata: payload.metadata
            }
          );
        }

        this.__handle( data.request, ( err, result, local_filenames ) => {
          Utils.removeFiles( local_filenames, () => {

            const response = {
              metadata: data.metadata
            };

            if ( err ) {
              response.error = err;
            }

            if ( result ) {
              response.result = result;
            }

            this.__respond( ch, msg, response );

          } );
        } );

      } );

    };
  }

  __handle( data, _done ) {

    const filenames = {
      source: this.__generateLocalFilename(),
      original: this.__generateLocalFilename(),
      channels: []
    };

    const done = function () {
      _done( ...arguments, Utils.listObjectValues( filenames ) );
    };

    this.log.debug( 'Downloading source.' );
    this.cloudStorage.download( data.audio.uri, filenames.source, ( err, src_url ) => {

      if ( err ) {
        return done( err, null );
      }

      this.log.debug( 'Transcoding original.' );
      Utils.transcode( filenames.source, filenames.original, {}, err => {

        if ( err ) {
          return done( err, null );
        }

        this.log.debug( 'Probing original.' );
        Utils.probe( filenames.original, ( err, probe ) => {

          if ( err ) {
            return done( err, null );
          }

          const metadata = probe.streams[ 0 ];
          const dst_base_url = CloudStorage.extendPathname( data.outputPath.uri, src_url.pathname );

          const links = {
            recordings: {
              original: CloudStorage.extendPathname( dst_base_url, 'original' ),
              channels: []
            },
            metadata: CloudStorage.extendPathname( dst_base_url, 'metadata' )
          };

          const transcodes = [];

          const uploads = [

            // Original transcoded file
            [
              filenames.original,
              links.recordings.original
            ],

            // Metadata file
            [
              Buffer.from( JSON.stringify( metadata ) ),
              links.metadata,
              {
                gzip: true,
                metadata: {
                  contentType: 'application/json'
                }
              }
            ]

          ];

          if ( metadata.channels > 1 ) {

            filenames.mixed = this.__generateLocalFilename();
            links.recordings.mixed = CloudStorage.extendPathname( dst_base_url, 'mixed' );

            transcodes.push( [ filenames.source, filenames.mixed, { mix: true } ] );
            uploads.push( [ filenames.mixed, links.recordings.mixed ] );

            for ( let i = 0; i < metadata.channels; ++i ) {

              const filename = this.__generateLocalFilename();
              const link = CloudStorage.extendPathname( dst_base_url, `channel.${i}` );

              filenames.channels.push( filename );
              links.recordings.channels.push( link );

              transcodes.push( [ filenames.source, filename, { channels: [ i ] } ] );
              uploads.push( [ filename, link ] );

            }

          }
          else {
            links.recordings.mixed = links.recordings.original;
            links.recordings.channels.push( links.recordings.original );
          }

          this.log.debug( 'Transcoding channels.' );

          Utils.bulk(
            Utils.transcode,
            transcodes,
            err => {

              if ( err ) {
                return done( err, null );
              }

              this.log.debug( 'Uploading.' );

              Utils.bulk(
                this.cloudStorage.upload.bind( this.cloudStorage ),
                uploads,
                err => {

                  if ( err ) {
                    return done( err, null );
                  }

                  done( null, { links: links } );

                }
              );

            }
          );

        } );

      } );

    } );

  }

  __respond( ch, msg, response ) {

    const error = response.error;

    // Log errors
    if ( error ) {

      error.data = msg.content.toString();

      if ( error.code >= 500 ) {
        this.log.warn( error );
        return ch.reject( msg, false );
      }

      if ( [ 401, 403, 409 ].indexOf( error.code ) !== -1 ) {
        this.log.error( error, true );
        return ch.reject( msg, false );
      }

      // Handle rate limits
      if ( error.code === 429 ) {
        this.log.warn( error, true );
        return ch.reject( msg, false );
      }

      this.log.error( error, !( error.code >= 400 ) );

      delete error.data;

    }

    // Publish response to response queue
    ch.publish(
      this.worker.exchange,
      this.worker.keys.response,
      Buffer.from( JSON.stringify( response ) ),
      {
        persistent: true,
        priority: msg.properties.priority || 0
      }
    );

    // Ack request
    ch.ack( msg );

  }

  __generateLocalFilename() {
    return path.resolve( this.worker.paths.storage, uuid.v4() );
  }

}

module.exports = TranscodeWorker;

if ( require.main === module ) {
  new TranscodeWorker().start( err => {
    if ( err ) {
      process.exit( 1 );
    }
  } );
}

process.on( 'SIGINT', () => {
  console.log( '[SIGINT] Terminating.' );
  process.exit( 0 );
} );

process.on( 'SIGTERM', () => {
  console.log( '[SIGTERM] Terminating.' );
  process.exit( 0 );
} );
