'use strict';

const cp = require( 'child_process' );
const fs = require( 'fs' );
const ffmpeg = require( 'fluent-ffmpeg' );

const DEFAULT_CODEC = 'flac';
const DEFAULT_FORMAT = 'flac';
const DEFAULT_SAMPLE_RATE = 8000;

module.exports = {
  bulk: bulk,
  removeFiles: removeFiles,
  rmdir: rmdir,
  probe: ffmpeg.ffprobe,
  transcode: transcode,
  listObjectValues: listObjectValues
};

function bulk( fn, args_list, done ) {
  (function nextUpload( i, n ) {

    if ( i >= n ) {
      return done( null );
    }

    fn( ...args_list[ i ], err => {

      if ( err ) {
        return done( err );
      }

      nextUpload( i + 1, n );

    } );

  })( 0, args_list.length );
}

function removeFiles( files, done ) {

  // If files is not an array, convert it into a singleton array
  if ( !Array.isArray( files ) ) {
    files = [ files ];
  }

  // Ensure there is something to remove
  if ( !files.length ) {
    return done();
  }

  const file = files.shift();

  if ( !file ) {
    return removeFiles( files, done );
  }

  // Remove the first file
  fs.unlink( file, () => removeFiles( files, done ) );

}

function rmdir( dir, done ) {
  cp.exec( `rm -rf ${dir}`, ( err, stdout, stderr ) => {

    if ( !( err || stderr ) ) {
      return done( null );
    }

    done(
      {
        reason: err && err.message || stderr,
        message: 'Unable to clear storage directory.'
      }
    );

  } );
}

function transcode( input, output, options, done ) {

  let finished = false;

  options = options || {};

  const params = [];

  params.push( '-c:a', options.codec || DEFAULT_CODEC );
  params.push( '-f', options.format || DEFAULT_FORMAT );
  params.push( '-ar', options.sampleRate || DEFAULT_SAMPLE_RATE );

  if ( options.channels ) {

    if ( !Array.isArray( options.channels ) ) {
      return done( new Error( 'Field options.channels must be an array.' ) );
    }

    options.channels.forEach( c => params.push( '-map_channel', '0.0.' + parseInt( c, 10 ) ) );

  }

  // Mix into single channel
  if ( options.mix ) {
    params.push( '-ac', 1 );
  }

  ffmpeg()
    .input( input )
    .output( output )
    .outputOptions( params )
    // .on( 'start', console.log )
    .on( 'error', err => {
      finish(
        {
          code: 400,
          reason: err.message,
          message: 'Unable to transcode audio.'
        }
      )
    } )
    .on( 'end', () => finish( null ) )
    .run();

  function finish() {
    if ( !finished ) {
      finished = true;
      done.apply( null, arguments );
    }
  }

}

function listObjectValues( o, vals = [] ) {

  Object.keys( o ).forEach( k => {

    const v = o[ k ];

    if ( typeof v !== 'object' ) {
      return vals.push( v );
    }

    listObjectValues( v, vals );

  } );

  return vals;

}
